import os
import sys

import pytest
from tox.config import parseconfig
from tox_direct.hookimpls import has_direct_envs

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path


class TestArgs:
    def test_normal(self, newconfig):
        newconfig("")
        print(os.getcwd())
        config = parseconfig([])
        assert not config.option.direct
        assert not config.option.direct_yolo

    def test_direct(self, newconfig):
        newconfig("")
        config = parseconfig(["--direct"])
        assert config.option.direct
        assert not config.option.direct_yolo

    def test_direct_yolo(self, newconfig):
        newconfig("")
        config = parseconfig(["--direct-yolo"])
        assert not config.option.direct
        assert config.option.direct_yolo


@pytest.mark.parametrize(
    "envlist, expectation",
    (
        ([], False),
        (["direct"], True),
        (["directwhatever"], True),
        (["whateverdirect"], True),
        (["whatdirectever"], True),
        (["direct", "another-direct"], True),
        (["normal", "another-normal"], False),
    ),
)
def test_has_direct_envs(envlist, expectation):
    if isinstance(expectation, bool):
        assert has_direct_envs(envlist) == expectation
    else:
        with pytest.raises(expectation):
            has_direct_envs(envlist)


def test_config(newconfig):
    config = newconfig(
        """
        [tox]
        skip_install = True
        [testenv:direct]
        skip_install = True
        deps = a
        """
    )
    direct = config.envconfigs["direct"]
    assert direct.deps == []
    assert direct.skip_install is True
    assert direct.basepython == sys.executable
    assert direct.envpython == sys.executable
    assert direct.get_envbindir() == str(Path(sys.executable).parent)


def test_does_not_interfere_with_single_normal_env(cmd, initproj):
    initproj(
        "does_not_interfer_with_normal_operation",
        filedefs={
            "tox.ini": """
                    [testenv:normal]
                    deps = decorator
                    commands = 
                        pip list
                        python -c 'import sys; print(sys.executable);'
            """
        },
    )
    r = cmd()
    assert r.ret == 0
    assert not r.session.config.option.direct
    assert not r.session.config.option.direct_yolo
    assert "We need direct action" not in r.out
    assert "won't build a package" not in r.out
    assert "won't install dependencies" not in r.out
    assert "won't install project" not in r.out
    assert "decorator" in r.out
    assert "congratulations :)" in r.out
    assert sys.executable not in r.out


def test_mixed_run_crashes_when_normal_env_needs_package(cmd, initproj):
    projectName = "example_project-1.3"
    initproj(
        projectName,
        filedefs={
            "tox.ini": """
                    [testenv:direct]
                    deps = decorator
                    commands = 
                        pip list
                        python -c 'import sys; print(sys.executable);'

                    [testenv:normal]
                    deps = decorator
                    commands = 
                        pip list
                        python -c 'import sys; print(sys.executable);'
            """
        },
    )
    r = cmd()
    assert r.ret == 1
    assert "[tox-direct] FATAL: tox env 'normal' needs a package" in r.err


def test_mixed_run_does_not_crash_when_normal_env_needs_no_package(cmd, initproj):
    projectName = "example_project-1.3"
    initproj(
        projectName,
        filedefs={
            "tox.ini": """
                    [testenv:direct]
                    deps = decorator
                    commands = 
                        pip list
                        python -c 'import sys; print(sys.executable);'

                    [testenv:normal]
                    usedevelop = True
                    deps = decorator
                    commands = 
                        pip list
                        python -c 'import sys; print(sys.executable);'
            """
        },
    )
    r = cmd()
    assert r.ret == 0
    assert "example-project 1.3" in r.out


def test_direct_vertical(cmd, initproj):
    initproj(
        "direct_vertical",
        filedefs={
            "tox.ini": """
                    [testenv:direct]
                    deps = dontcare
                    commands = python -c 'import sys; print(sys.executable);'
            """
        },
    )
    r = cmd("tox", "-vv")
    assert r.ret == 0
    assert not r.session.config.option.direct
    assert not r.session.config.option.direct_yolo
    assert "won't build a package" in r.out
    assert "won't install dependencies" in r.out
    assert "won't install project" in r.out
    assert "creating no virtual environment" in r.out
    assert "congratulations :)" in r.out
    assert sys.executable in r.out


def test_direct_yolo_normal_vertical(cmd, initproj):
    initproj(
        "yolo_normal_vertical",
        filedefs={
            "tox.ini": """
                    [testenv:normal]
                    deps = decorator
                    commands = 
                        pip list
                        python -c 'import sys; print(sys.executable);'
                        
            """
        },
    )
    r = cmd("tox", "-vv", "--direct-yolo")
    r.out = "\n".join(r.outlines)
    assert "YOLO!1!!" in r.out
    assert "decorator" in r.out
    assert sys.executable in r.out
    package = list((Path(".tox") / "dist").glob("*"))[0]
    assert "yolo_normal_vertical-0.1.zip" in package.name



from __future__ import print_function
import os
import sys

import py
import tox
from tox.session import reporter
from tox.exception import Error


class DIRECT:
    MARKER = "direct"
    ENV_VAR = "TOX_DIRECT"
    ENV_VAR_YOLO = "TOX_DIRECT_YOLO"
    SKIPSDIST_ORIGINAL = "_TOX_DIRECT_SKIPSDIST_ORIGINAL"
    PYTHON = py.path.local(sys.executable)


@tox.hookimpl
def tox_addoption(parser):
    parser.add_argument(
        "--direct",
        action="store_true",
        help="[tox-direct] deactivate venv, packaging and install steps - "
        "run commands directly "
        "(can also be achieved by setting {})".format(DIRECT.ENV_VAR),
    )
    parser.add_argument(
        "--direct-yolo",
        action="store_true",
        help="[tox-direct] do everything in host environment that would otherwise "
        "happen in an isolated virtual environment (can also be achieved "
        "by setting {} env var".format(DIRECT.ENV_VAR_YOLO),
    )


@tox.hookimpl
def tox_configure(config):
    if DIRECT.ENV_VAR in os.environ:
        config.option.direct = True
    if DIRECT.ENV_VAR_YOLO in os.environ:
        config.option.direct_yolo = True
    YOLO = config.option.direct_yolo
    if is_direct_run(config):
        if YOLO:
            reporter.info("YOLO! Do everything in the host environment.")
        setattr(config, DIRECT.SKIPSDIST_ORIGINAL, config.skipsdist)
        if not config.skipsdist and not YOLO:
            reporter.info("[tox-direct] won't build a package")
            config.skipsdist = True
        for name, envconfig in config.envconfigs.items():
            if not is_direct_call(config) and not is_direct_env(name):
                continue
            # TODO this could also be basepython on request (needed?)
            envconfig.get_envbindir = lambda: py.path.local(DIRECT.PYTHON.dirname)
            envconfig.get_envpython = lambda: py.path.local(DIRECT.PYTHON)
            assert envconfig.envpython == DIRECT.PYTHON, envconfig.envpython
            if envconfig.deps and not YOLO:
                reporter.info(
                    "[tox-direct] won't install dependencies in '{}'".format(name)
                )
                envconfig.deps = []
            if not envconfig.skip_install and not YOLO:
                envconfig.skip_install = True
                reporter.info("[tox-direct] won't install project in {}".format(name))


@tox.hookimpl
def tox_testenv_create(venv):
    if not is_direct_run(venv.envconfig.config):
        return  # normal behaviour
    isDirectCall = is_direct_call(venv.envconfig.config)
    isDirectVenv = is_direct_env(venv.name)
    YOLO = venv.envconfig.config.option.direct_yolo
    if not isDirectVenv and not YOLO:
        # direct run only safe for "normal" env if package not used in testenv
        needsPackage = (
            not venv.envconfig.skip_install
            and not venv.envconfig.usedevelop
        )
        if needsPackage and not getattr(venv.envconfig.config, DIRECT.SKIPSDIST_ORIGINAL):
            raise NormalEnvNeedsPackage(
                "[tox-direct] FATAL: tox env '{}' needs a package.\n"
                "Do not run this env as part of a direct run or "
                "run everything in the host (including package build) by running "
                "with --direct-yolo flag.\n"
                "WARNING: this will change the host environment.".format(venv.name)
            )

    if isDirectCall or isDirectVenv:
        venv.is_allowed_external = lambda _: True  # everything goes!
        reporter.info(
            "[tox-direct] creating no virtual environment - use:"
            " {}".format(venv.envconfig.envpython)
        )
        if not YOLO:
            return True
        reporter.info("[tox-direct] YOLO!1!!")


def is_direct_run(config):
    return is_direct_call(config) or has_direct_envs(config.envlist)


def is_direct_call(config):
    return config.option.direct or config.option.direct_yolo


def has_direct_envs(envlist):
    return any(is_direct_env(envname) for envname in envlist)


def is_direct_env(name):
    return DIRECT.MARKER in name


class NormalEnvNeedsPackage(Error):
    """Raised when a direct run contains a normal env needing a package."""



from setuptools import setup, find_packages

setup(
    name="tox-direct",
    description="plugin for tox - run everything directly (tox creates no virtual env)",
    version="0.3.4",
    packages=find_packages("src"),
    package_dir={"": "src"},
    entry_points={"tox": ["direct = tox_direct.hookimpls"]},
    install_requires=["tox>=3.12,<4", "py", "pathlib2"],
    extras_require={"test": ["pytest"], "publish": ["twine"]},
)
